#!/usr/bin/env python

from array import array
from itertools import combinations
import PIL.Image
import sys

def scale(v):
    return (v + 64) / 128

def point(rgb):
    return tuple(map(scale, rgb))

def index(p):
    return p[0] * 9 + p[1] * 3 + p[2]

def colour(i):
    return (255 * (i % 2), 255 * ((i / 2) % 2), 255 * ((i / 4) % 2))

def add(d, v):
    d[v] = (d.has_key(v) and d[v] or 0) + 1

def by_frequency(d):
    l = [(f, t) for (t, f) in d.items()]
    l.sort(reverse=True)
    return [i[1] for i in l]

def match(b, bases):
    return b in bases and b

def fallback(bases):
    for b in by_frequency(bases):
        if b not in ["_", "W"]:
            return b
    return by_frequency(bases)[0]

tones = [
    "___", "_BB", "BBB", # 00x
    "_GG", "_CC", "BCC", # 01x
    "GGG", "GCC", "CCC", # 02x
    "_RR", "_MM", "BMM", # 10x
    "_YY", "_*W", "BBW", # 11x
    "GYY", "GGW", "CCW", # 12x
    "RRR", "RMM", "MMM", # 20x
    "RYY", "RRW", "MMW", # 21x
    "YYY", "YYW", "WWW", # 22x
    ]

colours = ["_", "R", "G", "Y", "B", "M", "C", "W"]

if __name__ == "__main__":
    width = 320
    input_filename, output_filename = sys.argv[1:3]

    im = PIL.Image.open(input_filename)
    w, h = im.size
    height = (width * h) / w
    im = im.resize((width, height))

    usage = []
    base_usage = []
    toned = []

    for row in range(0, height):
        u = {}
        usage.append(u)
        bu = {}
        base_usage.append(bu)
        tr = []
        toned.append(tr)
        for column in range(0, width):
            rgb = im.getpixel((column, row))
            p = point(rgb)
            i = index(p)
            t = tones[i]
            add(u, t)
            if t[0] != "*":
                add(bu, t[0])
            if t[1] != "*":
                add(bu, t[1])
            if t[2] != "*":
                add(bu, t[2])
            tr.append(t)

    chosen = []

    for row, (u, bu) in enumerate(zip(usage, base_usage)):
        light = row % 2
        best = 0
        best_bases = None
        best_missing = None
        best_map = None
        for bases in combinations(bu, 4):
            bases = dict([(base, bu[base]) for base in bases])
            count = 0
            missing = []
            tone_map = {}
            for tone, freq in u.items():
                base = match(tone[1], bases) or match(light and tone[2] or tone[0], bases)
                if base:
                    tone_map[tone] = base
                    count += freq
                else:
                    missing.append(tone)
            if count > best:
                best_bases = bases
                best_missing = missing
                best_map = tone_map
                best = count
        chosen.append((best, best_bases or bases, best_map or tone_map, best_missing or missing))

    output = []

    for row, (tr, ch) in enumerate(zip(toned, chosen)):
        o = []
        for column, t in enumerate(tr):
            best, bases, tone_map, missing = ch
            base = tone_map.get(t) or fallback(bases)
            o.append(base)
            i = colours.index(base)
            im.putpixel((column, row), colour(i))

        output.append("".join(o))

    im.save(output_filename)

# vim: tabstop=4 expandtab shiftwidth=4
